unit UFastLists;

interface

uses
  Windows, Messages, SysUtils, Classes;

type

{ TStringList class }
  PFStringItem = ^TFStringItem;
  TFStringItem = record
    FString: string;
    FObject: TObject;
    FHash: integer;
  end;

  PFStringItemList = ^TFStringItemList;
  TFStringItemList = array[0..MaxListSize] of TFStringItem;

  TFStringList = class(TStrings)
  private
    FList: PFStringItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FDuplicates: TDuplicates;
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer);
    procedure InsertItem(Index: Integer; const S: string);
    procedure SetSorted(Value: Boolean);
    function CompareItem(item1, item2: PFStringItem): integer;
  protected
    function Get(Index: Integer): string; override;
    function GetCapacity: Integer; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetCapacity(NewCapacity: Integer); override;
  public
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: string; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: string): Integer; override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Sort; virtual;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted write SetSorted;
  end;

implementation

uses
  Consts, HyperStr;

{ TFStringList }

destructor TFStringList.Destroy;
begin
  inherited Destroy;
  if FCount <> 0 then Finalize(FList^[0], FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TFStringList.Add(const S: string): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(S, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: Error(SDuplicateString, 0);
      end;
  InsertItem(Result, S);
end;

procedure TFStringList.Clear;
begin
  if FCount <> 0 then
  begin
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
  end;
end;

procedure TFStringList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Finalize(FList^[Index]);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TFStringItem));
end;

procedure TFStringList.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(SListIndexError, Index2);
  ExchangeItems(Index1, Index2);
end;

procedure TFStringList.ExchangeItems(Index1, Index2: Integer);
var
  Temp: Integer;
  Item1, Item2: PFStringItem;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];
  Temp := Integer(Item1^.FString);
  Integer(Item1^.FString) := Integer(Item2^.FString);
  Integer(Item2^.FString) := Temp;
  Temp := Integer(Item1^.FObject);
  Integer(Item1^.FObject) := Integer(Item2^.FObject);
  Integer(Item2^.FObject) := Temp;
  Temp := Item1^.FHash;
  Item1^.FHash := Item2^.FHash;
  Item2^.FHash := Temp;
end;

function TFStringList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
  iHash: integer;
begin
  Result := False;
  iHash := Hash(S);
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    if iHash < FList^[I].FHash then C := 1
    else
      if iHash > FList^[I].FHash then C := -1
      else
        C := CompareText(FList^[I].FString, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

function TFStringList.Get(Index: Integer): string;
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList^[Index].FString;
end;

function TFStringList.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

function TFStringList.GetCount: Integer;
begin
  Result := FCount;
end;

function TFStringList.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList^[Index].FObject;
end;

procedure TFStringList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 512 then Delta := FCapacity div 4 else
    if FCapacity > 64 then Delta := 128 else
      if FCapacity > 8 then Delta := 16 else
        Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TFStringList.IndexOf(const S: string): Integer;
begin
  if not Sorted then Result := inherited IndexOf(S) else
    if not Find(S, Result) then Result := -1;
end;

procedure TFStringList.Insert(Index: Integer; const S: string);
begin
  if Sorted then Error(SSortedListError, 0);
  if (Index < 0) or (Index > FCount) then Error(SListIndexError, Index);
  InsertItem(Index, S);
end;

procedure TFStringList.InsertItem(Index: Integer; const S: string);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TFStringItem));
  with FList^[Index] do
  begin
    Pointer(FString) := nil;
    FObject := nil;
    FString := S;
    FHash := Hash(S);
  end;
  Inc(FCount);
end;

procedure TFStringList.Put(Index: Integer; const S: string);
begin
  if Sorted then Error(SSortedListError, 0);
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  FList^[Index].FString := S;
  FList^[Index].FHash := Hash(S);
end;

procedure TFStringList.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  FList^[Index].FObject := AObject;
end;

procedure TFStringList.QuickSort(L, R: Integer);
var
  I, J: Integer;
  tmpItem: PFStringItem;
begin
  repeat
    I := L;
    J := R;
    tmpItem := @FList^[(L + R) shr 1];
    repeat
      while CompareItem(@FList^[I], tmpItem) < 0 do inc(I);
      while CompareItem(@FList^[J], tmpItem) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

procedure TFStringList.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FList, NewCapacity * SizeOf(TFStringItem));
  FCapacity := NewCapacity;
end;

procedure TFStringList.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TFStringList.Sort;
begin
  if not Sorted and (FCount > 1) then
  begin
    QuickSort(0, FCount - 1);
  end;
end;

function TFStringList.CompareItem(item1, item2: PFStringItem): integer;
begin
  if item1^.FHash < item2^.FHash then Result := -1
  else
    if item1^.FHash > item2^.FHash then Result := 1
    else
      Result := CompareText(item1^.FString, item2^.FString);
end;

end.

